from kirin import types
from kirin.dialects import py, scf, ilist
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.ir.nodes.stmt import Statement


class ToRangeFor(RewriteRule):
    """Rewrite for-loops over `IList` iterables into for-loops over the range of their length.

    For example, rewrites:

    ```python
    for ele in ilist_values:
        ...

    ```

    to

    ```python
    for i in range(len(ilist_values)):
        ele = ilist_values[i]
        ...

    ```
    """

    def rewrite_Statement(self, node: Statement) -> RewriteResult:
        if not (
            isinstance(node, scf.For)
            and (iterable := node.iterable).type.is_subseteq(ilist.IListType)
            and not iterable.type.is_structurally_equal(types.Bottom)
        ):
            return RewriteResult()

        body_block = node.body.blocks[0]

        ele_arg = body_block.args[0]
        index = body_block.args.insert_from(0, types.Int)
        (ele_getitem := py.GetItem(iterable, index)).insert_before(
            body_block.first_stmt
        )
        ele_getitem.result.name = ele_arg.name
        ele_arg.replace_by(ele_getitem.result)
        body_block.args.delete(ele_arg)

        (len_stmt := py.Len(node.iterable)).insert_before(node)
        (zero := py.Constant(0)).insert_before(node)
        (one := py.Constant(1)).insert_before(node)
        (
            range_stmt := ilist.stmts.Range(zero.result, len_stmt.result, one.result)
        ).insert_before(node)

        node.iterable = range_stmt.result

        return RewriteResult(has_done_something=True)



from typing import Any

from kirin import types
from kirin.prelude import basic
from kirin.dialects import ilist


def test_method_union_binop_hint():

    @basic
    def main(x: ilist.IList[float, Any] | list[float]) -> float:
        return x[0]

    main.print()

    tps = main.arg_types

    assert len(tps) == 1
    assert tps[0] == types.Union(
        [ilist.IListType[types.Float, types.Any], types.List[types.Float]]
    )


def test_method_union_multi_hint():

    @basic
    def main(x: str | float | int):
        return x

    main.print()

    tps = main.arg_types

    assert len(tps) == 1
    assert tps[0] == types.Union([types.String, types.Float, types.Int])



# NOTE: this module is only interface, will be used inside
# the `ir` module try to minimize the dependencies as much
# as possible

from __future__ import annotations

import ast
import inspect
from abc import ABC
from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeAlias
from dataclasses import dataclass

from kirin.ir.attrs import types
from kirin.lowering.abc import Result
from kirin.lowering.exception import BuildError

if TYPE_CHECKING:
    from kirin.lowering.state import State


LoweringTransform: TypeAlias = Callable[[Any, "State[ast.AST]", ast.Call], Result]


@dataclass
class Transform:
    objs: tuple[Callable, ...]
    func: LoweringTransform


@dataclass
class akin:
    obj: Callable

    def __call__(
        self,
        func: LoweringTransform | Transform,
    ) -> Transform:
        if isinstance(func, Transform):
            return Transform((self.obj,) + func.objs, func.func)

        func_name = getattr(func, "__name__", "")
        if not func_name.startswith("lower_Call_"):
            raise SyntaxError(
                "lowering function should be prefixed with lower_Call_"
                f" but got {func_name}"
            )
        return Transform((self.obj,), func)


@dataclass
class FromPythonAST(ABC):
    callee_table: ClassVar[dict[object, Transform]]
    """a table of lowering transforms for ast.Call based
    on the callable object if avaiable as a global value.
    """

    def __init_subclass__(cls) -> None:
        # init the subclass first
        super().__init_subclass__()
        cls.callee_table = {}
        for _, value in inspect.getmembers(cls):
            if isinstance(value, Transform):
                for obj in value.objs:
                    cls.callee_table[obj] = value

    @property
    def names(self) -> list[str]:  # show the name without lower_
        return [name[6:] for name in dir(self) if name.startswith("lower_")]

    def lower(self, state: State[ast.AST], node: ast.AST) -> Result:
        """Entry point of dialect specific lowering."""
        return getattr(self, f"lower_{node.__class__.__name__}", self.unreachable)(
            state, node
        )

    def unreachable(self, state: State[ast.AST], node: ast.AST) -> Result:
        raise BuildError(f"unreachable reached for {node.__class__.__name__}")

    @staticmethod
    def _flatten_hint_binop(node: ast.expr) -> list[ast.expr]:
        """Flatten a binary operation tree into a list of expressions.

        This is useful for handling union types represented as binary operations.
        """
        hints = []

        def _recurse(n: ast.expr):
            if isinstance(n, ast.BinOp):
                _recurse(n.left)
                _recurse(n.right)
            else:
                hints.append(n)

        _recurse(node)
        return hints

    @staticmethod
    def get_hint(state: State[ast.AST], node: ast.expr | None) -> types.TypeAttribute:
        if node is None:
            return types.AnyType()

        # deal with union syntax
        if isinstance(node, ast.BinOp):
            hint_nodes = FromPythonAST._flatten_hint_binop(node)
            hint_ts = []
            for i in range(len(hint_nodes)):
                hint_ts.append(
                    FromPythonAST.get_hint(
                        state,
                        hint_nodes[i],
                    )
                )
            return types.Union(hint_ts)

        try:
            t = state.get_global(node).data

            return types.hint2type(t)
        except Exception as e:  # noqa: E722
            raise BuildError(f"expect a type hint, got {ast.unparse(node)}") from e


class NoSpecialLowering(FromPythonAST):
    pass



from kirin import ir
from kirin.prelude import basic
from kirin.dialects import ilist


def test_lambda_comp_with_closure():
    @basic(fold=False)
    def main(z, r):
        return (lambda x: x + z)(r)

    assert main(3, 4) == 7


def test_lambda_comp():
    @basic(fold=False)
    def main(z):
        return lambda x: x + z

    x = main(3)
    assert isinstance(x, ir.Method)
    assert x(4) == 7


def test_invoke_from_lambda_comp():

    @basic
    def foo(a):
        return a * 2

    @basic(fold=False)
    def main(z):
        return lambda x: x + foo(z)

    x = main(3)

    assert isinstance(x, ir.Method)
    assert x(4) == 10


def test_lambda_in_lambda():

    @basic(fold=False)
    def main(z):

        def my_foo(a):
            return lambda x: x * a

        return my_foo(z)

    x = main(3)

    assert isinstance(x, ir.Method)
    assert x(4) == 12


def test_ilist_map():

    @basic(fold=False)
    def main(z):
        return ilist.map(lambda x: x + z, ilist.range(10))

    x = main(3)
    assert len(x) == 10
    assert x.data == [3, 4, 5, 6, 7, 8, 9, 10, 11, 12]



from __future__ import annotations

import types
import inspect
from types import ModuleType
from typing import (
    TYPE_CHECKING,
    Union,
    Generic,
    TypeVar,
    Callable,
    ParamSpec,
    Concatenate,
    cast,
    overload,
)
from functools import update_wrapper
from dataclasses import field, dataclass
from collections.abc import Iterable

from typing_extensions import Self

from kirin.ir.method import Method
from kirin.ir.traits import SymbolTable, SymbolOpInterface
from kirin.ir.exception import CompilerError, ValidationError
from kirin.serialization.jsonserializer import get_json_serializer

if TYPE_CHECKING:
    from kirin.ir import Method, Dialect, Statement
    from kirin.lowering import Python
    from kirin.registry import Registry
    from kirin.serialization.base.serializer import Serializer
    from kirin.serialization.base.deserializer import Deserializer
    from kirin.serialization.core.serializationunit import SerializationUnit
    from kirin.serialization.core.serializationmodule import SerializationModule

PassParams = ParamSpec("PassParams")
RunPass = Callable[Concatenate[Method, PassParams], None]
RunPassGen = Callable[["DialectGroup"], RunPass[PassParams]]


@dataclass(init=False)
class DialectGroup(Generic[PassParams]):
    # method wrapper params
    Param = ParamSpec("Param")
    RetType = TypeVar("RetType")
    MethodTransform = Callable[[Callable[Param, RetType]], Method[Param, RetType]]

    data: frozenset["Dialect"]
    """The set of dialects in the group."""
    # NOTE: this is used to create new dialect groups from existing one
    run_pass_gen: RunPassGen[PassParams] | None = None
    """the function that generates the `run_pass` function.

    This is used to create new dialect groups from existing ones, while
    keeping the same `run_pass` function.
    """
    run_pass: RunPass[PassParams] | None = None
    """the function that runs the passes on the method."""

    lowering: Python = field(init=False, repr=False)
    """the lowering object used to lower the method."""

    symbol_table: dict[str, Statement] = field(
        default_factory=dict, init=False, repr=False
    )

    def __init__(
        self,
        dialects: Iterable[Union["Dialect", ModuleType]],
        run_pass: RunPassGen[PassParams] | None = None,
    ):
        self.symbol_table = {}
        self.data = frozenset(self.map_module(dialect) for dialect in dialects)
        if run_pass is None:
            self.run_pass_gen = None
            self.run_pass = None
        else:
            self.run_pass_gen = run_pass
            self.run_pass = run_pass(self)

        from kirin.lowering import Python

        self.lowering = Python(self)

    def __iter__(self):
        return iter(self.data)

    def __repr__(self) -> str:
        names = ", ".join(each.name for each in self.data)
        return f"DialectGroup([{names}])"

    @staticmethod
    def map_module(dialect: Union["Dialect", ModuleType]) -> "Dialect":
        """map the module to the dialect if it is a module.
        It assumes that the module has a `dialect` attribute
        that is an instance of [`Dialect`][kirin.ir.Dialect].
        """
        if isinstance(dialect, ModuleType):
            return getattr(dialect, "dialect")
        return dialect

    def add(self, dialect: Union["Dialect", ModuleType]) -> "DialectGroup":
        """add a dialect to the group.

        Args:
            dialect (Union[Dialect, ModuleType]): the dialect to add

        Returns:
            DialectGroup: the new dialect group with the added
        """
        return self.union([dialect])

    def union(self, dialect: Iterable[Union["Dialect", ModuleType]]) -> "DialectGroup":
        """union a set of dialects to the group.

        Args:
            dialect (Iterable[Union[Dialect, ModuleType]]): the dialects to union

        Returns:
            DialectGroup: the new dialect group with the union.
        """
        return DialectGroup(
            dialects=self.data.union(frozenset(self.map_module(d) for d in dialect)),
            run_pass=self.run_pass_gen,  # pass the run_pass_gen function
        )

    def discard(self, dialect: Union["Dialect", ModuleType]) -> "DialectGroup":
        """discard a dialect from the group.

        !!! note
            This does not raise an error if the dialect is not in the group.

        Args:
            dialect (Union[Dialect, ModuleType]): the dialect to discard

        Returns:
            DialectGroup: the new dialect group with the discarded dialect.
        """
        dialect_ = self.map_module(dialect)
        return DialectGroup(
            dialects=frozenset(
                each for each in self.data if each.name != dialect_.name
            ),
            run_pass=self.run_pass_gen,  # pass the run_pass_gen function
        )

    def __contains__(self, dialect) -> bool:
        """check if the dialect is in the group.

        Args:
            dialect (Union[Dialect, ModuleType]): the dialect to check.

        Returns:
            bool: True if the dialect is in the group, False otherwise.
        """
        return self.map_module(dialect) in self.data

    @property
    def registry(self) -> "Registry":
        """return the registry for the dialect group. This
        returns a proxy object that can be used to select
        the lowering interpreters, interpreters, and codegen
        for the dialects in the group.

        Returns:
            Registry: the registry object.
        """
        from kirin.registry import Registry

        return Registry(self)

    @overload
    def __call__(
        self,
        py_func: Callable[Param, RetType],
        *args: PassParams.args,
        **options: PassParams.kwargs,
    ) -> Method[Param, RetType]: ...

    @overload
    def __call__(
        self,
        py_func: None = None,
        *args: PassParams.args,
        **options: PassParams.kwargs,
    ) -> MethodTransform[Param, RetType]: ...

    def __call__(
        self,
        py_func: Callable[Param, RetType] | None = None,
        *args: PassParams.args,
        **options: PassParams.kwargs,
    ) -> Method[Param, RetType] | MethodTransform[Param, RetType]:
        """create a method from the python function.

        Args:
            py_func (Callable): the python function to create the method from.
            args (PassParams.args): the arguments to pass to the run_pass function.
            options (PassParams.kwargs): the keyword arguments to pass to the run_pass function.

        Returns:
            Method: the method created from the python function.
        """
        frame = inspect.currentframe()

        def wrapper(py_func: Callable) -> Method:
            assert isinstance(
                py_func, types.FunctionType
            ), f"Expected a function, got {type(py_func)}"
            if py_func.__name__ == "<lambda>":
                raise ValueError("Cannot compile lambda functions")

            lineno_offset, file = 0, ""
            mt = None
            if frame and frame.f_back is not None:
                call_site_frame = frame.f_back
                if py_func.__name__ in call_site_frame.f_locals:
                    mt = call_site_frame.f_locals[py_func.__name__]
                    if not isinstance(mt, Method):
                        raise CompilerError(
                            f"`{py_func.__name__}` is already defined in the current scope and is not a Method."
                        )

                lineno_offset = py_func.__code__.co_firstlineno - 1
                file = call_site_frame.f_code.co_filename

            code = self.lowering.python_function(py_func, lineno_offset=lineno_offset)
            arg_names = ["#self#"] + inspect.getfullargspec(py_func).args

            if mt:
                mt.mod = inspect.getmodule(py_func)
                mt.dialects = self
                mt.code = code
                mt.py_func = py_func
                mt.nargs = len(arg_names)
                mt.arg_names = arg_names
                mt.sym_name = py_func.__name__
                mt.file = file
                mt.lineno_begin = lineno_offset
                mt.run_passes = self.run_pass
                mt.update_backedges()  # update the callee
                self.recompile_callers(mt)
            else:
                mt = Method(
                    dialects=self,
                    code=code,
                    nargs=len(arg_names),
                    mod=inspect.getmodule(py_func),
                    py_func=py_func,
                    sym_name=py_func.__name__,
                    arg_names=arg_names,
                    file=file,
                    lineno_begin=lineno_offset,
                )

            if doc := inspect.getdoc(py_func):
                mt.__doc__ = doc

            def run_pass(mt: Method) -> None:
                if self.run_pass is not None:
                    try:
                        self.run_pass(mt, *args, **options)
                    except ValidationError as e:
                        e.attach(mt)
                        raise e

            mt.run_passes = run_pass
            run_pass(mt)
            self.update_symbol_table(mt)
            return mt

        if py_func is not None:
            return wrapper(py_func)
        return wrapper

    def recompile_callers(self, method: Method) -> None:
        for caller in method.backedges:
            if caller.run_passes:
                caller.run_passes(caller)
            # propagate the changes to all callers
            caller.dialects.recompile_callers(caller)
        return

    def update_symbol_table(self, method: Method) -> None:
        trait = method.code.get_trait(SymbolTable)
        if trait is None:
            return

        for stmt in method.code.walk():
            trait = stmt.get_trait(SymbolOpInterface)
            if trait is None:
                continue

            name = trait.get_sym_name(stmt).unwrap()
            if name in self.symbol_table:
                raise ValidationError(
                    stmt,
                    f"duplicate symbol `{name}` in dialect group",
                )
            self.symbol_table[name] = stmt

    def is_structurally_equal(
        self, other: DialectGroup, context: dict | None = None
    ) -> bool:
        if not isinstance(other, DialectGroup):
            return False
        if len(self.data) != len(other.data):
            return False
        x = sorted(self.data, key=lambda d: d.name)
        y = sorted(other.data, key=lambda d: d.name)
        for dialect1, dialect2 in zip(x, y):
            if dialect1.name != dialect2.name:
                return False
        return True

    def serialize(self, serializer: "Serializer") -> "SerializationUnit":
        return serializer.serialize_dialect_group(self)

    @classmethod
    def deserialize(
        cls: type[Self], serUnit: "SerializationUnit", deserializer: "Deserializer"
    ) -> Self:
        return cast(Self, deserializer.deserialize_dialect_group(serUnit))

    def encode(self, program) -> "SerializationModule":
        from kirin.serialization.base.serializer import Serializer

        serializer = Serializer()
        return serializer.encode(program)

    def decode(self, encoded: "SerializationModule") -> Method:
        from kirin.serialization.base.deserializer import Deserializer

        deserializer = Deserializer(dialect_group=self)
        return deserializer.decode(encoded)

    def encode_json(self, program: Method) -> str:
        encoded_module = self.encode(program)
        return get_json_serializer().encode(encoded_module)

    def decode_json(self, json_str: str) -> Method:
        decoded_module = get_json_serializer().decode(json_str)
        return self.decode(decoded_module)


def dialect_group(
    dialects: Iterable[Union["Dialect", ModuleType]],
) -> Callable[[RunPassGen[PassParams]], DialectGroup[PassParams]]:
    """Create a dialect group from the given dialects based on the
    definition of `run_pass` function.

    Args:
        dialects (Iterable[Union[Dialect, ModuleType]]): the dialects to include in the group.

    Returns:
        Callable[[RunPassGen[PassParams]], DialectGroup[PassParams]]: the dialect group.

    Example:
        ```python
        from kirin.dialects import cf, fcf, func, math

        @dialect_group([cf, fcf, func, math])
        def basic_no_opt(self):
            # initializations
            def run_pass(mt: Method) -> None:
                # how passes are applied to the method
                pass

            return run_pass
        ```
    """

    # NOTE: do not alias the annotation below
    def wrapper(
        transform: RunPassGen[PassParams],
    ) -> DialectGroup[PassParams]:
        ret = DialectGroup(dialects, run_pass=transform)
        update_wrapper(ret, transform)
        return ret

    return wrapper
